"""check for meaningful validation errors for various invalid input"""
from bioimageio.spec.shared import yaml


def test_list_instead_of_nested_schema(unet2d_nuclei_broad_latest):
    from bioimageio.spec.commands import validate

    assert yaml is not None
    data = yaml.load(unet2d_nuclei_broad_latest)

    # set wrong run_mode (list)
    data["run_mode"] = [{"name": "something"}]

    error = validate(data)["error"]
    assert isinstance(error, dict)
    assert len(error) == 1
    assert error["run_mode"] == ["Expected dictionary, but got list."]


def test_forward_compatibility_error(unet2d_fixed_shape):
    from bioimageio.spec.commands import validate

    assert yaml is not None
    data = yaml.load(unet2d_fixed_shape)

    data["authors"] = 42  # make sure rdf is invalid
    data["format_version"] = "9999.0.0"  # assume it is valid in a future format version

    error = validate(data)["error"]

    # even though the format version is correctly formatted, it should be mentioned here as we treat the future format
    # version as the current latest. If this attempted forward compatibility fails we have to report that we did it.
    assert isinstance(error, dict)
    assert "format_version" in error



import platform
from pathlib import Path

import pytest

from bioimageio.spec.dataset.raw_nodes import Dataset
from bioimageio.spec.shared import yaml


def test_spec_round_trip(unet2d_nuclei_broad_any_minor):
    from bioimageio.spec import load_raw_resource_description, serialize_raw_resource_description_to_dict

    assert yaml is not None
    expected = yaml.load(unet2d_nuclei_broad_any_minor)
    # monkeypatch: yaml.load already converts timestamp to datetime.datetime, while we serialize it to ISO 8601
    if "timestamp" in expected:
        expected["timestamp"] = expected["timestamp"].isoformat()

    # round-trip
    raw_model = load_raw_resource_description(unet2d_nuclei_broad_any_minor)
    serialized = serialize_raw_resource_description_to_dict(raw_model, convert_absolute_paths=True)

    assert isinstance(serialized, dict)
    assert expected == serialized

    # as we converted absolute paths back to relative, we need to set the root path
    serialized["root_path"] = unet2d_nuclei_broad_any_minor.parent
    raw_model_from_serialized = load_raw_resource_description(serialized)
    assert raw_model_from_serialized == raw_model


def test_spec_round_trip_w_attachments(unet2d_nuclei_broad_latest):
    from bioimageio.spec import load_raw_resource_description, serialize_raw_resource_description_to_dict

    assert yaml is not None
    data = yaml.load(unet2d_nuclei_broad_latest)
    data["root_path"] = unet2d_nuclei_broad_latest.parent

    # monkeypatch: yaml.load already converts timestamp to datetime.datetime, while we serialize it to ISO 8601
    if "timestamp" in data:
        data["timestamp"] = data["timestamp"].isoformat()

    data["attachments"] = {"files": ["some_file.ext"], "another_unknown_attachment": ["sub", "whatever", {"weird": 10}]}

    raw_model = load_raw_resource_description(data)

    serialized = serialize_raw_resource_description_to_dict(raw_model, convert_absolute_paths=True)
    assert isinstance(serialized, dict)
    serialized["root_path"] = unet2d_nuclei_broad_latest.parent
    assert serialized == data

    raw_model_from_serialized = load_raw_resource_description(serialized)
    assert raw_model_from_serialized == raw_model


def test_dataset_rdf_round_trip(dataset_rdf):
    from bioimageio.spec import load_raw_resource_description, serialize_raw_resource_description_to_dict

    assert yaml is not None
    data = yaml.load(dataset_rdf)
    raw = load_raw_resource_description(data)
    serialized = serialize_raw_resource_description_to_dict(raw)
    assert data == serialized


# todo: fix test on windows
@pytest.mark.skipif(
    platform.system() == "Windows", reason="OSError: [WinError 1314] A required privilege is not held by the client"
)
def test_serialize_with_link_in_path(dataset_rdf, tmp_path: Path):
    from bioimageio.spec import load_raw_resource_description, serialize_raw_resource_description_to_dict

    data = load_raw_resource_description(dataset_rdf)
    assert isinstance(data, Dataset)
    true_root = tmp_path / "root"
    true_root.mkdir()
    linked_root = tmp_path / "link"
    linked_root.symlink_to(true_root, target_is_directory=True)

    doc_path = linked_root / "docs.md"
    doc_path.write_text("# Documentation")

    data.root_path = true_root
    data.documentation = doc_path  # doc path only in root through link

    serialized = serialize_raw_resource_description_to_dict(data, convert_absolute_paths=True)
    assert serialized["documentation"] == "docs.md"


# todo: fix test on windows
@pytest.mark.skipif(
    platform.system() == "Windows", reason="OSError: [WinError 1314] A required privilege is not held by the client"
)
def test_serialize_with_link_in_root(dataset_rdf, tmp_path: Path):
    from bioimageio.spec import load_raw_resource_description, serialize_raw_resource_description_to_dict

    data = load_raw_resource_description(dataset_rdf)
    assert isinstance(data, Dataset)

    true_root = tmp_path / "root"
    true_root.mkdir()
    linked_root = tmp_path / "link"
    linked_root.symlink_to(true_root, target_is_directory=True)

    doc_path = true_root / "docs.md"
    doc_path.write_text("# Documentation")

    data.root_path = linked_root  # root path is symlink to true root
    data.documentation = doc_path

    serialized = serialize_raw_resource_description_to_dict(data, convert_absolute_paths=True)
    assert serialized["documentation"] == "docs.md"



import zipfile
from io import BytesIO, StringIO

from bioimageio.spec import (
    load_raw_resource_description,
    serialize_raw_resource_description,
    serialize_raw_resource_description_to_dict,
)
from bioimageio.spec.model import format_version, raw_nodes
from bioimageio.spec.shared import yaml


def test_validate_dataset(dataset_rdf):
    from bioimageio.spec.commands import validate

    summary = validate(dataset_rdf, update_format=True, update_format_inner=False)
    assert summary["status"] == "passed", summary
    summary = validate(dataset_rdf, update_format=False, update_format_inner=False)
    assert summary["status"] == "passed", summary


def test_validate_model_as_dict(unet2d_nuclei_broad_any):
    from bioimageio.spec.commands import validate

    assert not validate(unet2d_nuclei_broad_any, update_format=True, update_format_inner=False)["error"]
    assert not validate(unet2d_nuclei_broad_any, update_format=False, update_format_inner=False)["error"]


def test_validate_model_as_url():
    from bioimageio.spec.commands import validate

    assert not validate(
        "https://raw.githubusercontent.com/bioimage-io/spec-bioimage-io/main/example_specs/models/unet2d_nuclei_broad/rdf.yaml",
        update_format=True,
        update_format_inner=False,
    )["error"]
    assert not validate(
        "https://raw.githubusercontent.com/bioimage-io/spec-bioimage-io/main/example_specs/models/unet2d_nuclei_broad/rdf.yaml",
        update_format=False,
        update_format_inner=False,
    )["error"]


def test_validate_model_as_zenodo_sandbox_doi():
    from bioimageio.spec.commands import validate

    doi = "10.5281/zenodo.5744489"
    assert not validate(doi, update_format=False, update_format_inner=False)["error"]


def test_validate_model_as_zenodo_doi():
    from bioimageio.spec.commands import validate

    doi = "10.5281/zenodo.5744490"
    assert not validate(doi, update_format=False, update_format_inner=False)["error"]

    # expecting UnconvertibleError due to missing sha256
    assert validate(doi, update_format=True, update_format_inner=False)["error"]


def test_validate_model_as_bioimageio_full_version_id_partner():
    from bioimageio.spec.commands import validate

    full_version_id = "ilastik/isbi2012_neuron_segmentation_challenge/latest"
    summary = validate(full_version_id, update_format=False, update_format_inner=False)
    assert summary["status"] == "passed", summary["error"]


def test_validate_model_as_bioimageio_full_version_id_zenodo():
    from bioimageio.spec.commands import validate

    full_version_id = "10.5281/zenodo.5874741/5874742"
    summary = validate(full_version_id, update_format=False, update_format_inner=False)
    assert summary["status"] == "passed", summary["error"]


def test_validate_model_as_bioimageio_resource_id_partner():
    from bioimageio.spec.commands import validate

    resource_id = "ilastik/isbi2012_neuron_segmentation_challenge"
    summary = validate(resource_id, update_format=False, update_format_inner=False)
    assert summary["status"] == "passed", summary["error"]


def test_validate_model_as_bioimageio_resource_id_zenodo():
    from bioimageio.spec.commands import validate

    resource_id = "10.5281/zenodo.5874741"
    summary = validate(resource_id, update_format=False, update_format_inner=False)
    assert summary["status"] == "passed", summary["error"]


def test_validate_model_as_bytes_io(unet2d_nuclei_broad_latest):
    from bioimageio.spec.commands import validate

    # load from path and serialize with absolute paths
    raw_rd = load_raw_resource_description(unet2d_nuclei_broad_latest)
    data_str = serialize_raw_resource_description(raw_rd, convert_absolute_paths=False)

    data = BytesIO(data_str.encode("utf-8"))
    data.seek(0)
    assert not validate(data, update_format=True, update_format_inner=False)["error"]
    data.seek(0)
    assert not validate(data, update_format=False, update_format_inner=False)["error"]


def test_validate_model_as_string_io(unet2d_nuclei_broad_latest):
    from bioimageio.spec.commands import validate

    # load from path and serialize with absolute paths
    raw_rd = load_raw_resource_description(unet2d_nuclei_broad_latest)
    data_str = serialize_raw_resource_description(raw_rd, convert_absolute_paths=False)

    data = StringIO(data_str)
    data.seek(0)
    assert not validate(data, update_format=True, update_format_inner=False)["error"]
    data.seek(0)
    assert not validate(data, update_format=False, update_format_inner=False)["error"]


def test_validate_model_as_bytes(unet2d_nuclei_broad_latest):
    from bioimageio.spec.commands import validate

    # load from path and serialize with absolute paths
    raw_rd = load_raw_resource_description(unet2d_nuclei_broad_latest)
    data_str = serialize_raw_resource_description(raw_rd, convert_absolute_paths=False)

    data = data_str.encode("utf-8")
    assert not validate(data, update_format=True, update_format_inner=False)["error"]
    assert not validate(data, update_format=False, update_format_inner=False)["error"]


def test_validate_model_as_string(unet2d_nuclei_broad_latest):
    from bioimageio.spec.commands import validate

    # load from path and serialize with absolute paths
    raw_rd = load_raw_resource_description(unet2d_nuclei_broad_latest)
    data = serialize_raw_resource_description(raw_rd, convert_absolute_paths=False)

    assert not validate(data, update_format=True, update_format_inner=False)["error"]
    assert not validate(data, update_format=False, update_format_inner=False)["error"]


def test_validate_model_package_as_bytes(unet2d_nuclei_broad_latest):
    from bioimageio.spec.commands import validate

    # load from path and serialize with absolute paths
    raw_rd = load_raw_resource_description(unet2d_nuclei_broad_latest)
    rdf_str = serialize_raw_resource_description(raw_rd, convert_absolute_paths=False)

    data = BytesIO()
    with zipfile.ZipFile(data, "w") as zf:
        zf.writestr("rdf.yaml", rdf_str)

    data.seek(0)
    assert not validate(data, update_format=True, update_format_inner=False)["error"]
    data.seek(0)
    assert not validate(data, update_format=False, update_format_inner=False)["error"]


def test_validate_model_package_on_disk(unet2d_nuclei_broad_latest, tmpdir):
    from bioimageio.spec.commands import validate

    # load from path and serialize with absolute paths
    raw_rd = load_raw_resource_description(unet2d_nuclei_broad_latest)
    rdf_str = serialize_raw_resource_description(raw_rd, convert_absolute_paths=False)

    zf_path = tmpdir / "package.zip"
    with zipfile.ZipFile(zf_path, "w") as zf:
        zf.writestr("rdf.yaml", rdf_str)

    assert not validate(zf_path, update_format=True, update_format_inner=False)["error"]
    assert not validate(zf_path, update_format=False, update_format_inner=False)["error"]


def test_validate_invalid_model(unet2d_nuclei_broad_latest):
    from bioimageio.spec.commands import validate

    # load from path and serialize with absolute paths
    raw_rd = load_raw_resource_description(unet2d_nuclei_broad_latest)
    data = serialize_raw_resource_description_to_dict(raw_rd)

    del data["test_inputs"]  # invalidate data
    assert validate(data, update_format=True, update_format_inner=False)["error"]
    assert validate(data, update_format=False, update_format_inner=False)["error"]


def test_validate_generates_warnings(unet2d_nuclei_broad_latest):
    from bioimageio.spec.commands import validate

    raw_rd = load_raw_resource_description(unet2d_nuclei_broad_latest)
    data = serialize_raw_resource_description_to_dict(raw_rd)
    data["license"] = "BSD-2-Clause-FreeBSD"
    data["run_mode"] = {"name": "fancy"}
    summary = validate(data, update_format=True, update_format_inner=False)

    assert summary["warnings"]


def test_update_format(unet2d_nuclei_broad_before_latest, tmp_path):
    from bioimageio.spec.commands import update_format

    path = tmp_path / "rdf_new.yaml"
    update_format(unet2d_nuclei_broad_before_latest, path)

    assert path.exists()
    model = load_raw_resource_description(path)
    assert model.format_version == format_version


def test_update_rdf_using_paths(unet2d_nuclei_broad_latest, tmp_path):
    from bioimageio.spec.commands import update_rdf

    in_path = unet2d_nuclei_broad_latest
    assert in_path.exists()
    update_path = tmp_path / "update.yaml"
    assert yaml is not None
    yaml.dump(dict(name="updated", outputs=[{"name": "updated", "halo": ["KEEP", "DROP", 0, 9, 9]}]), update_path)
    out_path = tmp_path / "output.yaml"
    update_rdf(in_path, update_path, out_path)
    actual = yaml.load(out_path)
    assert actual["name"] == "updated"
    assert actual["outputs"][0]["name"] == "updated"
    assert actual["outputs"][0]["halo"] == [0, 0, 9, 9]


def test_update_rdf_using_dicts(unet2d_nuclei_broad_latest):
    from bioimageio.spec.commands import update_rdf

    # load from path and serialize with absolute paths
    raw_rd = load_raw_resource_description(unet2d_nuclei_broad_latest)
    source = serialize_raw_resource_description_to_dict(raw_rd, convert_absolute_paths=False)

    update = dict(name="updated", outputs=[{"name": "updated", "halo": ["KEEP", "DROP", 0, 9, 9]}])
    actual = update_rdf(source, update)
    assert isinstance(actual, dict)
    assert actual["name"] == "updated"
    assert actual["outputs"][0]["name"] == "updated"
    assert actual["outputs"][0]["halo"] == [0, 0, 9, 9]


def test_update_rdf_using_dicts_in_place(unet2d_nuclei_broad_latest):
    from bioimageio.spec.commands import update_rdf

    # load from path and serialize with absolute paths
    raw_rd = load_raw_resource_description(unet2d_nuclei_broad_latest)
    source = serialize_raw_resource_description_to_dict(raw_rd, convert_absolute_paths=False)

    update = dict(name="updated", outputs=[{"name": "updated", "halo": ["KEEP", "DROP", 0, 9, 9]}])
    update_rdf(source, update, output=source)
    actual = source
    assert actual["name"] == "updated"
    assert actual["outputs"][0]["name"] == "updated"
    assert actual["outputs"][0]["halo"] == [0, 0, 9, 9]


def test_update_rdf_using_rd(unet2d_nuclei_broad_latest):
    from bioimageio.spec.commands import update_rdf

    source = load_raw_resource_description(unet2d_nuclei_broad_latest)
    update = dict(name="updated", outputs=[{"name": "updated", "halo": ["KEEP", "DROP", 0, 9, 9]}])
    actual = update_rdf(source, update)
    assert isinstance(actual, raw_nodes.Model)
    assert actual.name == "updated"
    assert actual.outputs[0].name == "updated"
    assert actual.outputs[0].halo == [0, 0, 9, 9]



import pathlib
from datetime import datetime, timezone
from typing import Any, Dict

import numpy
import pytest
from marshmallow import Schema, ValidationError
from numpy.testing import assert_equal
from pytest import raises

from bioimageio.spec.model import schema
from bioimageio.spec.shared import fields, raw_nodes


class TestArray:
    def test_unequal_nesting_depth(self):
        with raises(ValidationError):
            fields.Array(fields.Integer(strict=True)).deserialize([[1, 2], 3])

    def test_uneuqal_sublen(self):
        with raises(ValidationError):
            fields.Array(fields.Integer(strict=True)).deserialize([[1, 2], [3]])

    def test_scalar(self):
        data = 1
        expected = data
        actual = fields.Array(fields.Integer(strict=True)).deserialize(data)
        assert_equal(actual, expected)

    def test_invalid_scalar(self):
        data = "invalid"
        with raises(ValidationError):
            fields.Array(fields.Integer(strict=True)).deserialize(data)

    def test_2d(self):
        data = [[1, 2], [3, 4]]
        expected = numpy.array(data, dtype=int)
        actual = fields.Array(fields.Integer(strict=True)).deserialize(data)
        assert_equal(actual, expected)

    def test_wrong_dtype(self):
        data = [[1, 2], [3, 4.5]]
        with raises(ValidationError):
            fields.Array(fields.Integer(strict=True)).deserialize(data)


class TestDateTime:
    def test_datetime_from_str(self):
        timestamp = "2019-12-11T12:22:32+00:00"
        expected = datetime.fromisoformat(timestamp)
        actual = fields.DateTime().deserialize(timestamp)
        assert expected == actual

    def test_datetime_from_datetime(self):
        expected = datetime.now()
        assert expected == fields.DateTime().deserialize(expected)

    def test_datetime_iso_with_zulu_offset(self):
        timestamp_non_zulu = "2019-12-11T12:22:32+00:00"
        timestamp_zulu = "2019-12-11T12:22:32Z"
        expected = datetime(2019, 12, 11, 12, 22, 32, tzinfo=timezone.utc)
        actual1 = fields.DateTime().deserialize(timestamp_non_zulu)
        actual2 = fields.DateTime().deserialize(timestamp_zulu)
        assert expected == actual1
        assert expected == actual2


class TestShape:
    def test_explicit_input_shape(self):
        data = [1, 2, 3]
        expected = data
        actual = fields.ExplicitShape().deserialize(data)
        assert expected == actual

    def test_explicit_output_shape(self):
        data = [1, 2, 3]
        expected = data
        actual = schema.OutputTensor().fields["shape"].deserialize(data)
        assert expected == actual

    def test_min_step_input_shape(self):
        data = {"min": [1, 2, 3], "step": [0, 1, 3]}
        expected = raw_nodes.ParametrizedInputShape(**data)
        actual = fields.Union(
            [fields.ExplicitShape(), fields.Nested(schema.ParametrizedInputShape())], required=True
        ).deserialize(data)
        assert actual == expected

    def test_output_shape(self):
        data: Dict[str, Any] = {"reference_tensor": "in1", "scale": [1, 2, 3], "offset": [0, 1, 3]}
        expected = raw_nodes.ImplicitOutputShape(**data)
        actual = fields.Union(
            [fields.ExplicitShape(), fields.Nested(schema.ImplicitOutputShape())], required=True
        ).deserialize(data)
        assert actual == expected


class TestURI:
    def test_missing_scheme_is_invalid(self):
        # local relative paths used to be valid "uris"
        relative_path = "relative_file/path.txt"

        with pytest.raises(ValidationError):
            fields.URI().deserialize(relative_path)


class TestUnion:
    def test_error_messages(self):
        union = fields.Union([fields.String(), fields.Number()])
        try:
            union.deserialize([1])
        except ValidationError as e:
            assert isinstance(e, ValidationError)
            assert len(e.messages) == 3, e.messages

    def test_union_with_absolute_path(self):
        class DummySchema(Schema):
            source = fields.Union([fields.URI(), fields.RelativeLocalPath()])  # we use this case in a lot of places

        s = DummySchema()
        data = dict(source="C:/repos")

        with pytest.raises(ValidationError):
            s.load(data)


class TestRelativeLocalPath:
    def test_simple_file_name(self):
        fname = "unet2d.py"
        expected = pathlib.Path(fname)
        actual = fields.RelativeLocalPath().deserialize(fname)
        assert isinstance(actual, pathlib.Path)
        assert actual == expected


class TestDependencies:
    class MinimalSchema(Schema):
        dep = fields.Dependencies()

    def test_url_roundtrip(self):
        manager = "conda"
        file = "https://raw.githubusercontent.com/bioimage-io/spec-bioimage-io/main/example_specs/models/unet2d_nuclei_broad/environment.yaml"
        dep_input = dict(dep=f"{manager}:{file}")
        s = self.MinimalSchema()

        node = s.load(dep_input)
        dep = node["dep"]
        assert isinstance(dep, raw_nodes.Dependencies)
        assert dep.manager == manager
        assert isinstance(dep.file, raw_nodes.URI)
        assert str(dep.file) == file

        dep_serialized = s.dump(node)
        assert dep_serialized == dep_input



from bioimageio.spec.shared.common import get_args


def test_weights_formats_have_raw_nodes():
    from bioimageio.spec.model import raw_nodes, schema

    weights_formats = [wf for wf in get_args(raw_nodes.WeightsFormat)]
    weights_entry_class_names = [wf.title().replace("_", "") + "WeightsEntry" for wf in weights_formats]

    # all defined weights formats need their schema and nodes implemented...
    for wecn in weights_entry_class_names:
        assert hasattr(schema, wecn), wecn
        assert hasattr(raw_nodes, wecn), wecn

    # every WeightEntry schema needs to validate its corresponding weights_format
    for wf, wecn in zip(weights_formats, weights_entry_class_names):
        comparable = getattr(schema, wecn)().fields["weights_format"].validate.comparable
        assert comparable == wf, (comparable, wf)



"""bioimageio.spec extensions for BioImage.IO community partners"""
# todo: consider moving to its own module, e.g. bioimageio.partner
# todo: if not moving to its own module, add the dependency to a 'partner' extra in setup.py



from dataclasses import asdict
from typing import Tuple

import pytest
from packaging.version import Version

from bioimageio.spec.model import schema
from bioimageio.spec.shared import yaml


def test_model_format_version_conversion(unet2d_nuclei_broad_before_latest, unet2d_nuclei_broad_latest):
    from bioimageio.spec.model.converters import maybe_convert

    assert yaml is not None
    old_model_data = yaml.load(unet2d_nuclei_broad_before_latest)
    model_data = yaml.load(unet2d_nuclei_broad_latest)

    expected = asdict(schema.Model().load(model_data))
    converted_data = maybe_convert(old_model_data)
    actual = asdict(schema.Model().load(converted_data))

    if old_model_data["format_version"] == "0.1.0":
        # expect converted description
        for ipt in expected["inputs"]:
            ipt["description"] = ipt["name"]

        for out in expected["outputs"]:
            out["description"] = out["name"]

    # ignore new optional fields
    if tuple(map(int, old_model_data["format_version"].split("."))) < (0, 4, 0):
        expected.pop("maintainers")
        actual.pop("maintainers")
    if tuple(map(int, old_model_data["format_version"].split("."))) < (0, 4, 4):
        expected.pop("download_url")
        actual.pop("download_url")
        expected.pop("training_data")
        actual.pop("training_data")

    for key, item in expected.items():
        assert key in actual, key
        assert actual[key] == item, key

    for key, item in actual.items():
        assert key in expected
        assert item == expected[key]


# todo: break forward compatibility on major version difference?
@pytest.mark.parametrize("v_diff", [(0, 0, 1), (0, 1, 0), (1, 0, 0), (0, 1, 1)])
def test_forward_compatible(v_diff: Tuple[int, int, int], unet2d_nuclei_broad_latest):
    from bioimageio.spec import load_raw_resource_description
    from bioimageio.spec.model import format_version

    fv_key = "format_version"

    assert yaml is not None
    model_data = yaml.load(unet2d_nuclei_broad_latest)

    v_latest: Version = Version(format_version)
    v_future: str = ".".join(
        [str(latest + diff) for latest, diff in zip([v_latest.major, v_latest.minor, v_latest.micro], v_diff)]
    )

    future_model_data = dict(model_data)
    future_model_data[fv_key] = v_future

    rd = load_raw_resource_description(future_model_data)
    assert rd.format_version == format_version
    assert hasattr(rd, "config")
    assert rd.config["bioimageio"]["original_format_version"] == v_future



import pathlib

from bioimageio.spec.model import raw_nodes
from bioimageio.spec import collection
from bioimageio.spec.shared import yaml


def test_load_raw_model(unet2d_nuclei_broad_any):
    from bioimageio.spec import load_raw_resource_description

    raw_model = load_raw_resource_description(unet2d_nuclei_broad_any)
    assert raw_model


def test_load_upsample_raw_model(upsamle_model_rdf):
    from bioimageio.spec import load_raw_resource_description

    raw_model = load_raw_resource_description(upsamle_model_rdf)
    assert raw_model


def test_loaded_remote_raw_model_is_valid(unet2d_nuclei_broad_url):
    from bioimageio.spec import load_raw_resource_description

    raw_model = load_raw_resource_description(unet2d_nuclei_broad_url)
    raw_model = load_raw_resource_description(raw_model)
    assert raw_model


def test_load_raw_model_fixed_shape(unet2d_fixed_shape):
    from bioimageio.spec import load_raw_resource_description

    raw_model = load_raw_resource_description(unet2d_fixed_shape)
    assert raw_model


def test_load_raw_model_diff_output_shape(unet2d_diff_output_shape):
    from bioimageio.spec import load_raw_resource_description

    raw_model = load_raw_resource_description(unet2d_diff_output_shape)
    assert raw_model


def test_load_raw_model_multi_tensor(unet2d_multi_tensor):
    from bioimageio.spec import load_raw_resource_description

    raw_model = load_raw_resource_description(unet2d_multi_tensor)
    assert raw_model


def test_load_raw_model_expanded_output_shape(unet2d_expanded_output_shape):
    from bioimageio.spec import load_raw_resource_description

    raw_model = load_raw_resource_description(unet2d_expanded_output_shape)
    assert raw_model


def test_load_raw_model_hpa(hpa_model):
    from bioimageio.spec import load_raw_resource_description

    raw_model = load_raw_resource_description(hpa_model)
    assert raw_model


def test_load_raw_model_stardist(stardist_model):
    from bioimageio.spec import load_raw_resource_description

    raw_model = load_raw_resource_description(stardist_model)
    assert raw_model


def test_load_raw_model_unet2d_keras_tf(unet2d_keras_tf):
    from bioimageio.spec import load_raw_resource_description

    raw_model = load_raw_resource_description(unet2d_keras_tf, update_to_format="latest")
    assert isinstance(raw_model, raw_nodes.Model)
    # test attachments
    assert len(raw_model.attachments.files) == 1
    attachment = raw_model.attachments.files[0]
    assert isinstance(attachment, pathlib.Path)
    assert (raw_model.root_path / attachment).exists()


def test_load_raw_model_unet2d_keras_tf2(unet2d_keras_tf2):
    from bioimageio.spec import load_raw_resource_description

    raw_model = load_raw_resource_description(unet2d_keras_tf2, update_to_format="latest")
    assert isinstance(raw_model, raw_nodes.Model)
    # test attachments
    assert len(raw_model.attachments.files) == 3
    attachments = raw_model.attachments.files
    assert all(isinstance(at, pathlib.Path) and (raw_model.root_path / at).exists() for at in attachments)


def test_load_raw_model_to_format(unet2d_nuclei_broad_before_latest):
    from bioimageio.spec import load_raw_resource_description

    assert yaml is not None
    data = yaml.load(unet2d_nuclei_broad_before_latest)
    data["root_path"] = unet2d_nuclei_broad_before_latest.parent
    format_targets = [(0, 3), (0, 4)]
    format_version = tuple(map(int, data["format_version"].split(".")[:2]))

    for target in format_targets:
        if format_version <= target:
            to_format = ".".join(map(str, target))
            raw_model = load_raw_resource_description(data, update_to_format=to_format)
            assert raw_model.format_version[: raw_model.format_version.rfind(".")] == to_format


def test_load_raw_model_converts_invalid_name(unet2d_nuclei_broad_base_path):
    from bioimageio.spec.model.raw_nodes import Model
    from bioimageio.spec import load_raw_resource_description

    assert yaml is not None
    model_dict = yaml.load(unet2d_nuclei_broad_base_path / "rdf_v0_4_0.yaml")
    model_dict["root_path"] = unet2d_nuclei_broad_base_path
    model_dict["name"] = "invalid/name"
    model = load_raw_resource_description(model_dict)
    assert isinstance(model, Model)
    assert model.name == "invalidname"


def test_collection_with_relative_path_in_rdf_source_of_an_entry(partner_collection):
    from bioimageio.spec import load_raw_resource_description
    from bioimageio.spec.collection.utils import resolve_collection_entries
    from bioimageio.spec.dataset.v0_2.raw_nodes import Dataset

    coll = load_raw_resource_description(partner_collection)
    assert isinstance(coll, collection.raw_nodes.Collection)
    resolved_entries = resolve_collection_entries(coll)
    for entry_rdf, entry_error in resolved_entries:
        assert isinstance(entry_rdf, Dataset)
        assert isinstance(entry_rdf.documentation, pathlib.Path) and entry_rdf.documentation.as_posix().endswith(
            "example_specs/collections/partner_collection/datasets/dummy-dataset/README.md"
        )



import pytest

from bioimageio.spec.shared import fields


def test_cite_entry():
    from bioimageio.spec.rdf.schema import CiteEntry

    data = {
        "text": "Title",
        "doi": "https://doi.org/10.1109/5.771073",
        "url": "https://ieeexplore.ieee.org/document/771073",
    }

    CiteEntry().load(data)


def test_cite_field_option1():
    """only way we allow to specify listed, nested schemas.

    Limitation to allow better exception and warning messages and make the code in general more concise.
    """
    from bioimageio.spec.rdf.schema import CiteEntry

    data = [
        {
            "text": "Title",
            "doi": "https://doi.org/10.1109/5.771073",
            "url": "https://ieeexplore.ieee.org/document/771073",
        }
    ] * 2

    cite_field = fields.List(fields.Nested(CiteEntry()), required=True)
    cite_field.deserialize(data)


# we (arbitrarily) don't allow this. Test for reference only. see fields.Nested for details
# def test_cite_field_option2():
#     from bioimageio.spec.rdf.schema import CiteEntry
#
#     data = [
#         {
#             "text": "Title",
#             "doi": "https://doi.org/10.1109/5.771073",
#             "url": "https://ieeexplore.ieee.org/document/771073",
#         }
#     ] * 2
#
#     cite_field = fields.Nested(CiteEntry(many=True), required=True)
#     out = cite_field.deserialize(data)
#     assert len(out) == 2


# we (arbitrarily) don't allow this. Test for reference only. see fields.Nested for details
# def test_cite_field_option3():
#     from bioimageio.spec.rdf.schema import CiteEntry
#
#     data = [
#         {
#             "text": "Title",
#             "doi": "https://doi.org/10.1109/5.771073",
#             "url": "https://ieeexplore.ieee.org/document/771073",
#         }
#     ] * 2
#
#     cite_field = fields.Nested(CiteEntry(many=True), many=True, required=True)
#     out = cite_field.deserialize(data)
#     assert len(out) == 2



from bioimageio.spec.shared import LICENSES


def test_licenses_read_and_reformatted():
    """Make sure LICENSES is dict of dicts"""
    assert isinstance(LICENSES, dict)
    assert all(isinstance(v, dict) for k, v in LICENSES.items())



from bioimageio.spec import load_raw_resource_description, serialize_raw_resource_description_to_dict


def test_validate(unet2d_nuclei_broad_collection):
    from bioimageio.spec.commands import validate

    assert not validate(unet2d_nuclei_broad_collection, update_format=True, update_format_inner=False)["error"]
    assert not validate(unet2d_nuclei_broad_collection, update_format=False, update_format_inner=False)["error"]


def test_validate_invalid(unet2d_nuclei_broad_collection):
    from bioimageio.spec.commands import validate

    raw_rd = load_raw_resource_description(unet2d_nuclei_broad_collection)
    data = serialize_raw_resource_description_to_dict(raw_rd, convert_absolute_paths=False)

    data["collection"][0]["name"] = 1  # invalidate data
    assert validate(data, update_format=True, update_format_inner=False)["error"]
    assert validate(data, update_format=False, update_format_inner=False)["error"]



import dataclasses
import pathlib
from datetime import datetime
from typing import Any, Dict

import pytest


def test_uri():
    from bioimageio.spec.shared.raw_nodes import URI

    uri_from_string = URI("https://john.doe@www.example.com:123/forum/questions/?tag=networking&order=newest#top")
    uri = URI(
        scheme="https",
        authority="john.doe@www.example.com:123",
        path="/forum/questions/",
        query="tag=networking&order=newest",
        fragment="top",
    )

    assert str(uri_from_string) == str(uri)
    assert uri_from_string == uri


def test_replace_uri_wo_uri_string():
    from bioimageio.spec.shared.raw_nodes import URI

    uri_string = "https://john.doe@www.example.com:123/forum/questions/?tag=networking&order=newest#top"
    uri_string = uri_string.replace("top", "bottom")

    uri = URI(
        scheme="https",
        authority="john.doe@www.example.com:123",
        path="/forum/questions/",
        query="tag=networking&order=newest",
        fragment="top",
    )
    uri = dataclasses.replace(uri, fragment="bottom")
    assert uri_string == str(uri)


def test_uri_is_relative_path():
    from bioimageio.spec.shared.raw_nodes import URI

    # todo: figure out if it is important to keep a trailing slash.
    #  atm uri_from_string removes it (using urllib.parse.urlparse)
    # uri_from_string = URI("file:forum/questions/")
    # uri = URI(scheme="file", path="forum/questions/")

    uri_from_string = URI("file:forum/questions")
    uri = URI(scheme="file", path="forum/questions")

    assert str(uri_from_string) == str(uri)
    assert uri_from_string == uri


def test_uri_is_url():
    from bioimageio.spec.shared.raw_nodes import URI

    url = "https://raw.githubusercontent.com/bioimage-io/spec-bioimage-io/main/example_specs/models/unet2d_nuclei_broad/test_input.npy"
    uri = URI(url)
    assert str(uri) == url


def test_uri_truediv():
    from bioimageio.spec.shared.raw_nodes import URI

    url = "https://raw.githubusercontent.com/bioimage-io/spec-bioimage-io/main/example_specs/models/unet2d_nuclei_broad?download=1"
    rel_path = "test_input.npy"
    expected = URI(
        f"https://raw.githubusercontent.com/bioimage-io/spec-bioimage-io/main/example_specs/models/unet2d_nuclei_broad/{rel_path}?download=1"
    )
    uri = URI(url)
    assert expected == uri / rel_path
    assert expected == uri / rel_path  # ensure we did not change uri in-place


def test_uri_parent():
    from bioimageio.spec.shared.raw_nodes import URI

    url = "https://raw.githubusercontent.com/bioimage-io/spec-bioimage-io/main/example_specs/models/unet2d_nuclei_broad/test_input.npy?download=1"
    expected = URI(
        "https://raw.githubusercontent.com/bioimage-io/spec-bioimage-io/main/example_specs/models/unet2d_nuclei_broad?download=1"
    )
    uri = URI(url)
    assert expected == uri.parent
    assert expected == uri.parent  # ensure we did not change uri in-place


def test_general_rdf_accepts_unknown_fields():
    from bioimageio.spec.rdf.raw_nodes import RDF

    rdf = RDF(
        format_version="0.2.0",
        name="test_rdf",
        authors=[],
        cite=[],
        description="description text",
        documentation=pathlib.Path("README.md"),
        links=[],
        tags=[],
        unknown_weird_test_field="shouldn't be here",
    )
    assert rdf.name == "test_rdf"


def test_model_does_not_accept_unknown_fields():
    from bioimageio.spec.model.raw_nodes import Model

    model_kwargs: Dict[str, Any] = dict(
        authors=[],
        cite=[],
        description="description text",
        documentation=pathlib.Path("README.md"),
        format_version="0.3.2",
        inputs=[],
        license="MIT",
        name="test_model",
        outputs=[],
        tags=[],
        test_inputs=[],
        test_outputs=[],
        timestamp=datetime.now(),
        weights={},
    )
    # check that model_kwargs are valid
    Model(**model_kwargs)
    with pytest.raises(TypeError):
        Model(**model_kwargs, unknown_weird_test_field="shouldn't be here")  # type: ignore



from datetime import datetime

import pytest
from marshmallow import ValidationError

from bioimageio.spec.model.v0_4 import raw_nodes as raw_nodes_m04
from bioimageio.spec.shared import yaml


def test_model_rdf_is_valid_general_rdf(unet2d_nuclei_broad_latest):
    from bioimageio.spec.rdf.schema import RDF

    assert yaml is not None
    data = yaml.load(unet2d_nuclei_broad_latest)
    data["root_path"] = unet2d_nuclei_broad_latest.parent

    RDF().load(data)


def test_model_does_not_accept_unknown_fields(unet2d_nuclei_broad_latest):
    from bioimageio.spec.model.schema import Model

    assert yaml is not None
    data = yaml.load(unet2d_nuclei_broad_latest)
    data["root_path"] = unet2d_nuclei_broad_latest.parent

    data["unknown_additional_field"] = "shouldn't be here"

    with pytest.raises(ValidationError):
        Model().load(data)


@pytest.fixture
def model_dict():
    """
    Valid model dict fixture
    """
    return {
        "documentation": "./docs.md",
        "license": "MIT",
        "git_repo": "https://github.com/bioimage-io/python-bioimage-io",
        "format_version": "0.4.0",
        "description": "description",
        "authors": [
            {"name": "Author 1", "affiliation": "Affiliation 1"},
            {"name": "Author 2", "affiliation": "Affiliation 2"},
        ],
        "maintainers": [
            {"name": "Author 1", "affiliation": "Affiliation 1", "github_user": "githubuser1"},
            {"name": "Author 2", "affiliation": "Affiliation 2", "github_user": "githubuser2"},
        ],
        "timestamp": datetime.now(),
        "cite": [{"text": "Paper title", "doi": "doi"}],
        "inputs": [
            {"name": "input_1", "description": "Input 1", "data_type": "float32", "axes": "xyc", "shape": [128, 128, 3]}
        ],
        "outputs": [
            {
                "name": "output_1",
                "description": "Output 1",
                "data_type": "float32",
                "axes": "xyc",
                "shape": [128, 128, 3],
            }
        ],
        "name": "Model",
        "tags": [],
        "weights": {},
        "test_inputs": ["test_ipt.npy"],
        "test_outputs": ["test_out.npy"],
        "type": "model",
    }


def test_model_schema_accepts_run_mode(model_dict):
    from bioimageio.spec.model.schema import Model

    model_schema = Model()
    model_dict.update({"run_mode": {"name": "special_run_mode", "kwargs": dict(marathon=True)}})
    validated_data = model_schema.load(model_dict)
    assert validated_data


@pytest.mark.parametrize(
    "format",
    ["pytorch_state_dict", "torchscript", "keras_hdf5", "tensorflow_js", "tensorflow_saved_model_bundle", "onnx"],
)
def test_model_schema_accepts_valid_weight_formats(model_dict, format):
    from bioimageio.spec.model.schema import Model

    model_schema = Model()
    model_dict.update({"weights": {format: {"source": "local_weights"}}})
    if format == "pytorch_state_dict":
        model_dict["weights"][format]["architecture"] = "file.py:Model"
        model_dict["weights"][format]["architecture_sha256"] = "0" * 64  # dummy sha256

    validated_data = model_schema.load(model_dict)
    assert validated_data


def test_model_schema_raises_invalid_name(model_dict):
    from bioimageio.spec.model.schema import Model

    model_schema = Model()
    model_dict["name"] = "invalid/name"
    with pytest.raises(ValidationError):
        model_schema.load(model_dict)


def test_model_schema_raises_invalid_input_name(model_dict):
    from bioimageio.spec.model.schema import Model

    model_schema = Model()
    model_dict["inputs"][0]["name"] = "invalid/name"
    with pytest.raises(ValidationError):
        model_schema.load(model_dict)


def test_model_schema_raises_invalid_output_name(model_dict):
    from bioimageio.spec.model.schema import Model

    model_schema = Model()
    model_dict["outputs"][0]["name"] = "invalid/name"
    with pytest.raises(ValidationError):
        model_schema.load(model_dict)


def test_model_0_4_raises_on_duplicate_tensor_names(invalid_rdf_v0_4_0_duplicate_tensor_names):
    from bioimageio.spec.model.schema import Model
    from bioimageio.spec.model.v0_3.schema import Model as Model_v03

    assert yaml is not None
    data = yaml.load(invalid_rdf_v0_4_0_duplicate_tensor_names)

    model_schema = Model()
    with pytest.raises(ValidationError):
        model_schema.load(data)

    # as 0.3 the model should still be valid with some small changes
    model_schema = Model_v03()
    data["format_version"] = "0.3.3"
    data["language"] = "python"
    data["framework"] = "pytorch"
    data["source"] = data["weights"]["pytorch_state_dict"].pop("architecture")
    data["kwargs"] = data["weights"]["pytorch_state_dict"].pop("kwargs")
    data["sha256"] = data["weights"]["pytorch_state_dict"].pop("architecture_sha256")

    valid_data = model_schema.load(data)
    assert valid_data


def test_output_fixed_shape_too_small(model_dict):
    from bioimageio.spec.model.schema import Model

    model_dict["outputs"] = [
        {
            "name": "output_1",
            "description": "Output 1",
            "data_type": "float32",
            "axes": "xyc",
            "shape": [128, 128, 3],
            "halo": [32, 128, 0],
        }
    ]

    with pytest.raises(ValidationError) as e:
        Model().load(model_dict)

    assert e.value.messages == {
        "_schema": ["Minimal shape [128 128   3] of output output_1 is too small for halo [32, 128, 0]."]
    }


def test_output_ref_shape_mismatch(model_dict):
    from bioimageio.spec.model.schema import Model

    model_dict["outputs"] = [
        {
            "name": "output_1",
            "description": "Output 1",
            "data_type": "float32",
            "axes": "xyc",
            "shape": {"reference_tensor": "input_1", "scale": [1, 2, 3, 4], "offset": [0, 0, 0, 0]},
        }
    ]

    with pytest.raises(ValidationError) as e:
        Model().load(model_dict)

    assert e.value.messages == {
        "_schema": [
            "Referenced tensor input_1 with 3 dimensions does not match output tensor output_1 with 4 dimensions."
        ]
    }


def test_output_ref_shape_too_small(model_dict):
    from bioimageio.spec.model.schema import Model

    model_dict["outputs"] = [
        {
            "name": "output_1",
            "description": "Output 1",
            "data_type": "float32",
            "axes": "xyc",
            "shape": {"reference_tensor": "input_1", "scale": [1, 2, 3], "offset": [0, 0, 0]},
            "halo": [256, 128, 0],
        }
    ]

    with pytest.raises(ValidationError) as e:
        Model().load(model_dict)

    assert e.value.messages == {
        "_schema": ["Minimal shape [128. 256.   9.] of output output_1 is too small for halo [256, 128, 0]."]
    }


def test_model_has_parent_with_uri(model_dict):
    from bioimageio.spec.model.schema import Model

    model_dict["parent"] = dict(uri="https://doi.org/10.5281/zenodo.5744489")

    valid_data = Model().load(model_dict)
    assert isinstance(valid_data, raw_nodes_m04.Model)


def test_model_has_parent_with_id(model_dict):
    from bioimageio.spec.model.schema import Model

    model_dict["parent"] = dict(id="10.5281/zenodo.5764892")

    valid_data = Model().load(model_dict)
    assert isinstance(valid_data, raw_nodes_m04.Model)


def test_model_with_expanded_output(model_dict):
    from bioimageio.spec.model.schema import Model

    model_dict["outputs"] = [
        {
            "name": "output_1",
            "description": "Output 1",
            "data_type": "float32",
            "axes": "xyzc",
            "shape": dict(
                scale=[1, 1, None, 1],
                offset=[0, 0, 7, 0],
                reference_tensor="input_1",
            ),
        }
    ]

    model = Model().load(model_dict)
    assert isinstance(model, raw_nodes_m04.Model)
    out0_shape = model.outputs[0].shape
    assert isinstance(out0_shape, raw_nodes_m04.ImplicitOutputShape)
    assert out0_shape.scale == [1, 1, None, 1]



import warnings
from pathlib import Path
from typing import Any, Dict, Union

from bioimageio.spec.shared import resolve_rdf_source
from .imjoy_plugin_parser import get_plugin_as_rdf  # type: ignore
from ..shared.raw_nodes import URI


def enrich_partial_rdf_with_imjoy_plugin(partial_rdf: Dict[str, Any], root: Union[URI, Path]) -> Dict[str, Any]:
    """
    a (partial) rdf may have 'rdf_resource' or 'source' which resolve to rdf data that may be overwritten.
    Due to resolving imjoy plugins this is not done in bioimageio.spec.collection atm
    """

    enriched_rdf = {}
    if "rdf_source" in partial_rdf:
        rdf_source = partial_rdf["rdf_source"]
        if isinstance(rdf_source, str) and rdf_source.split("?")[0].endswith(".imjoy.html"):
            # rdf_source is an imjoy plugin
            rdf_source = dict(get_plugin_as_rdf(rdf_source))

        else:
            # rdf_source is an actual rdf
            if not isinstance(rdf_source, dict):
                try:
                    rdf_source, rdf_source_name, rdf_source_root = resolve_rdf_source(rdf_source)
                except Exception as e:
                    try:
                        rdf_source, rdf_source_name, rdf_source_root = resolve_rdf_source(root / rdf_source)
                    except Exception as ee:
                        rdf_source = {}
                        warnings.warn(f"Failed to resolve `rdf_source`: 1. {e}\n2. {ee}")
                    else:
                        rdf_source["root_path"] = rdf_source_root  # enables remote source content to be resolved
                else:
                    rdf_source["root_path"] = rdf_source_root  # enables remote source content to be resolved

        assert isinstance(rdf_source, dict)
        enriched_rdf.update(rdf_source)

    if "source" in partial_rdf:
        if partial_rdf["source"].split("?")[0].endswith(".imjoy.html"):
            rdf_from_source = get_plugin_as_rdf(partial_rdf["source"])
            enriched_rdf.update(rdf_from_source)

    enriched_rdf.update(partial_rdf)  # initial partial rdf overwrites fields from rdf_source or source
    return enriched_rdf



import pathlib

from bioimageio.spec.shared import yaml


def test_get_resource_package_content(unet2d_nuclei_broad_latest, unet2d_nuclei_broad_url):
    from bioimageio.spec import get_resource_package_content

    from_local_content = get_resource_package_content(unet2d_nuclei_broad_latest)
    from_remote_content = get_resource_package_content(unet2d_nuclei_broad_url)
    local_keys = set(from_local_content)
    remote_keys = set(from_remote_content)
    assert local_keys == remote_keys


def test_load_animal_nickname():
    from bioimageio.spec import load_raw_resource_description
    from bioimageio.spec.model.v0_4.raw_nodes import Model as Model04

    nickname = "impartial-shrimp"
    model = load_raw_resource_description(nickname)
    assert isinstance(model, Model04)
    assert ".".join(model.format_version.split(".")[:2]) == "0.4"
    assert model.config["bioimageio"]["nickname"] == nickname


def test_resolve_download_url(unet2d_nuclei_broad_latest):
    from bioimageio.spec import load_raw_resource_description
    from bioimageio.spec.model.v0_4.raw_nodes import Model as Model04

    assert yaml is not None
    data = yaml.load(unet2d_nuclei_broad_latest)
    data["root_path"] = unet2d_nuclei_broad_latest.parent  # set root path manually as we load from the manipulated dict
    data["download_url"] = "relative_path_to.zip"

    model = load_raw_resource_description(data)
    assert isinstance(model, Model04)
    assert isinstance(model.download_url, pathlib.Path)



import pytest
from marshmallow import missing

from bioimageio.spec.model import raw_nodes, schema


def test_tensor_schema_preprocessing():
    data = {
        "name": "input_1",
        "description": "Input 1",
        "data_type": "float32",
        "axes": "xyc",
        "shape": [128, 128, 3],
        "preprocessing": [
            {
                "name": "scale_range",
                "kwargs": {"max_percentile": 99, "min_percentile": 5, "mode": "per_sample", "axes": "xy"},
            }
        ],
    }
    validated_data = schema.InputTensor().load(data)
    assert isinstance(validated_data, raw_nodes.InputTensor)
    assert validated_data.name == data["name"]
    assert validated_data.description == data["description"]
    assert validated_data.data_type == data["data_type"]
    assert validated_data.axes == data["axes"]
    assert validated_data.shape == data["shape"]

    assert isinstance(validated_data.preprocessing, list)
    assert len(validated_data.preprocessing) == 1
    preprocessing = validated_data.preprocessing[0]
    assert preprocessing.name == "scale_range"


@pytest.mark.parametrize(
    "data",
    [
        {"name": "input_1", "description": "Input 1", "data_type": "float32", "axes": "xyc", "shape": [128, 128, 3]},
        {"name": "input_1", "description": "Input 1", "data_type": "float32", "axes": "xyc", "shape": [128, 128, 3]},
    ],
)
def test_tensor_schema_no_preprocessing(data):
    validated_data = schema.InputTensor().load(data)
    assert validated_data.preprocessing is missing


@pytest.mark.parametrize("schema_instance", [schema.InputTensor(), schema.OutputTensor()])
def test_tensor_schema_optional_description(schema_instance):
    data = {"name": "tensor_1", "data_type": "float32", "axes": "xyc", "shape": [128, 128, 3]}
    validated_data = schema_instance.load(data)
    assert validated_data.description is missing



import sys
from pathlib import Path
from pprint import pprint
from typing import Any, Callable, Dict, Optional, Union

import typer

from bioimageio.spec import __version__, collection, commands, model, rdf
from bioimageio.spec.shared.raw_nodes import URI

enrich_partial_rdf_with_imjoy_plugin: Optional[Callable[[Dict[str, Any], Union[URI, Path]], Dict[str, Any]]]
try:
    from bioimageio.spec.partner.utils import enrich_partial_rdf_with_imjoy_plugin
except ImportError:
    enrich_partial_rdf_with_imjoy_plugin = None
    partner_help = ""
else:
    partner_help = f"\n+\nbioimageio.spec.partner {__version__}\nimplementing:\n\tpartner collection RDF {collection.format_version}"

help_version = (
    f"bioimageio.spec {__version__}"
    "\nimplementing:"
    f"\n\tcollection RDF {collection.format_version}"
    f"\n\tgeneral RDF {rdf.format_version}"
    f"\n\tmodel RDF {model.format_version}" + partner_help
)

# prevent rewrapping with \b\n: https://click.palletsprojects.com/en/7.x/documentation/#preventing-rewrapping
app = typer.Typer(
    help="\b\n" + help_version,
    context_settings={"help_option_names": ["-h", "--help", "--version"]},  # make --version display help with version
)  # https://typer.tiangolo.com/


@app.command()
def validate(
    rdf_source: str = typer.Argument(..., help="RDF source as relative file path or URI"),
    update_format: bool = typer.Option(
        False,
        help="Update format version to the latest (might fail even if source adheres to an old format version). "
        "To inform the format update the source may specify fields of future versions in "
        "config:future:<future version>.",  # todo: add future documentation
    ),
    update_format_inner: bool = typer.Option(
        None, help="For collection RDFs only. Defaults to value of 'update-format'."
    ),
    verbose: bool = typer.Option(False, help="show traceback of unexpected (no ValidationError) exceptions"),
):
    summary = commands.validate(rdf_source, update_format, update_format_inner)
    if summary["error"] is not None:
        print(f"Error in {summary['name']}:")
        pprint(summary["error"])
        if verbose:
            print("traceback:")
            pprint(summary["traceback"])
        ret_code = 1
    else:
        print(f"No validation errors for {summary['name']}")
        ret_code = 0

    if summary["warnings"]:
        print(f"Validation Warnings for {summary['name']}:")
        pprint(summary["warnings"])

    sys.exit(ret_code)


validate.__doc__ = commands.validate.__doc__


if enrich_partial_rdf_with_imjoy_plugin is not None:

    @app.command()
    def validate_partner_collection(
        rdf_source: str = typer.Argument(..., help="RDF source as relative file path or URI"),
        update_format: bool = typer.Option(
            False,
            help="Update format version to the latest (might fail even if source adheres to an old format version). "
            "To inform the format update the source may specify fields of future versions in "
            "config:future:<future version>.",  # todo: add future documentation
        ),
        update_format_inner: bool = typer.Option(
            None, help="For collection RDFs only. Defaults to value of 'update-format'."
        ),
        verbose: bool = typer.Option(False, help="show traceback of unexpected (no ValidationError) exceptions"),
    ):
        assert enrich_partial_rdf_with_imjoy_plugin is not None
        summary = commands.validate(
            rdf_source, update_format, update_format_inner, enrich_partial_rdf=enrich_partial_rdf_with_imjoy_plugin
        )
        if summary["error"] is not None:
            print(f"Error in {summary['name']}:")
            pprint(summary["error"])
            if verbose:
                print("traceback:")
                pprint(summary["traceback"])
            ret_code = 1
        else:
            print(f"No validation errors for {summary['name']}")
            ret_code = 0

        if summary["warnings"]:
            print(f"Validation Warnings for {summary['name']}:")
            pprint(summary["warnings"])

        sys.exit(ret_code)

    cmd_doc = commands.validate.__doc__
    assert cmd_doc is not None
    validate_partner_collection.__doc__ = (
        "A special version of the bioimageio validate command that enriches the RDFs defined in collections by parsing any "
        "associated imjoy plugins. This is implemented using the 'enrich_partial_rdf' of the regular validate command:\n"
        + cmd_doc
    )


@app.command()
def update_format(
    rdf_source: str = typer.Argument(..., help="RDF source as relative file path or URI"),
    path: str = typer.Argument(..., help="Path to save the RDF converted to the latest format"),
):
    try:
        commands.update_format(rdf_source, path)
        ret_code = 0
    except Exception as e:
        print(f"update-format failed with {e}")
        ret_code = 1
    sys.exit(ret_code)


update_format.__doc__ == commands.update_format.__doc__


@app.command()
def update_rdf(
    source: str = typer.Argument(..., help="relative file path or URI to RDF source"),
    update: str = typer.Argument(..., help="relative file path or URI to (partial) RDF as update"),
    output: Path = typer.Argument(..., help="Path to save the updated RDF to"),
    validate: bool = typer.Option(True, help="Whether or not to validate the updated RDF"),
):
    """Update a given RDF with a (partial) RDF-like update"""
    try:
        commands.update_rdf(source, update, output, validate)
        ret_code = 0
    except Exception as e:
        print(f"update-rdf failed with {e}")
        ret_code = 1
    sys.exit(ret_code)


@app.callback()
def callback():
    typer.echo(help_version)


if __name__ == "__main__":
    app()



from dataclasses import dataclass
from pathlib import Path
from typing import Any

import pytest
from packaging.version import Version

from bioimageio.spec.shared import raw_nodes
from bioimageio.spec.shared.node_transformer import NodeTransformer, NodeVisitor, iter_fields
from bioimageio.spec.shared.raw_nodes import ResourceDescription


@dataclass
class MyNode(raw_nodes.RawNode):
    field_a: str
    field_b: int


def test_iter_fields():
    entry = MyNode("a", 42)
    assert [("field_a", "a"), ("field_b", 42)] == list(iter_fields(entry))


@dataclass
class Content:
    data: str


class TestNodeVisitor:
    @dataclass
    class Tree(raw_nodes.RawNode):
        left: Any
        right: Any

    @dataclass
    class URL:
        url: str

    @pytest.fixture
    def tree(self):
        return self.Tree(
            self.Tree(self.Tree(None, None), self.URL("https://example.com")),
            self.Tree(None, self.Tree(None, self.Tree(None, None))),
        )

    def test_node(self, tree):
        visitor = NodeVisitor()
        visitor.visit(tree)

    def test_node_transform(self, tree):
        class MyTransformer(NodeTransformer):
            def transform_URL(self, node):
                return Content(f"content of url {node.url}")

        assert isinstance(tree.left.right, self.URL)
        transformer = MyTransformer()
        transformed_tree = transformer.transform(tree)
        assert isinstance(transformed_tree.left.right, Content)


def test_resolve_remote_relative_path():
    from bioimageio.spec.shared.node_transformer import RelativePathTransformer

    remote_rdf = raw_nodes.URI(
        "https://raw.githubusercontent.com/bioimage-io/spec-bioimage-io/main/example_specs/models/"
        "unet2d_nuclei_broad/rdf.yaml"
    )
    remote_relative_path = Path("unet2d.py")

    uri = RelativePathTransformer(root=remote_rdf.parent).transform(remote_relative_path)

    assert (
        str(uri) == "https://raw.githubusercontent.com/bioimage-io/spec-bioimage-io/main/example_specs/models/"
        "unet2d_nuclei_broad/unet2d.py"
    )


@pytest.mark.parametrize(
    "data_update_expected",
    [
        ({"a": 1, "b": 2}, {"a": {"c": 1}}, {"a": {"c": 1}, "b": 2}),
        ({"a": [0, 1, 2, 3]}, {"a": [5]}, {"a": [5, 1, 2, 3]}),
        ([0, 1, 2, 3], [5], [5, 1, 2, 3]),
        ([0, {"a": [1]}, 2, 3], ["DROP", {"a": ["KEEP", 2]}], [{"a": [1, 2]}, 2, 3]),
        (
            ResourceDescription(
                format_version="0.1.0",
                name="resource",
                type="test",
                version=Version("0.1.0"),
            ),
            {"name": "updated resource"},
            ResourceDescription(
                format_version="0.1.0",
                name="updated resource",
                type="test",
                version=Version("0.1.0"),
            ),
        ),
    ],
)
def test_update_nested(data_update_expected):
    from bioimageio.spec.shared import update_nested

    data, update, expected = data_update_expected
    actual = update_nested(data, update)
    assert actual == expected



import os
import subprocess
import zipfile
from typing import Sequence

from bioimageio.spec.io_ import (
    load_raw_resource_description,
    save_raw_resource_description,
    serialize_raw_resource_description,
)
from bioimageio.spec.shared import yaml


def run_subprocess(commands: Sequence[str], **kwargs) -> subprocess.CompletedProcess:
    return subprocess.run(commands, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, encoding="utf-8", **kwargs)


def test_cli_validate_model(unet2d_nuclei_broad_latest):
    ret = run_subprocess(["bioimageio", "validate", str(unet2d_nuclei_broad_latest)])
    assert ret.returncode == 0


def test_cli_validate_model_url():
    ret = run_subprocess(
        [
            "bioimageio",
            "validate",
            "https://raw.githubusercontent.com/bioimage-io/spec-bioimage-io/main/example_specs/models/unet2d_nuclei_broad/rdf.yaml",
        ]
    )
    assert ret.returncode == 0


def test_cli_validate_model_url_wo_cache():
    env = os.environ.copy()
    env["BIOIMAGEIO_USE_CACHE"] = "false"
    ret = run_subprocess(
        [
            "bioimageio",
            "validate",
            "https://raw.githubusercontent.com/bioimage-io/spec-bioimage-io/main/example_specs/models/unet2d_nuclei_broad/rdf.yaml",
        ],
        env=env,
    )
    assert ret.returncode == 0


def test_cli_validate_model_doi():
    ret = run_subprocess(["bioimageio", "validate", "10.5281/zenodo.5744489"])
    assert ret.returncode == 0


def test_cli_validate_model_package(unet2d_nuclei_broad_latest, tmpdir):
    zf_path = tmpdir / "package.zip"

    # load from path and serialize with absolute paths
    raw_rd = load_raw_resource_description(unet2d_nuclei_broad_latest)
    rdf_str = serialize_raw_resource_description(raw_rd, convert_absolute_paths=False)

    with zipfile.ZipFile(zf_path, "w") as zf:
        zf.writestr("rdf.yaml", rdf_str)

    ret = run_subprocess(["bioimageio", "validate", str(zf_path)])
    assert ret.returncode == 0


def test_cli_validate_model_package_wo_cache(unet2d_nuclei_broad_latest, tmpdir):
    env = os.environ.copy()
    env["BIOIMAGEIO_USE_CACHE"] = "false"

    # load from path and serialize with absolute paths
    raw_rd = load_raw_resource_description(unet2d_nuclei_broad_latest)
    rdf_str = serialize_raw_resource_description(raw_rd, convert_absolute_paths=False)

    zf_path = tmpdir / "package.zip"
    with zipfile.ZipFile(zf_path, "w") as zf:
        zf.writestr("rdf.yaml", rdf_str)

    ret = run_subprocess(["bioimageio", "validate", str(zf_path)], env=env)
    assert ret.returncode == 0


def test_cli_update_format(unet2d_nuclei_broad_before_latest, tmp_path):
    in_path = tmp_path / "rdf.yaml"
    save_raw_resource_description(load_raw_resource_description(unet2d_nuclei_broad_before_latest), in_path)
    assert in_path.exists()
    path = tmp_path / "rdf_new.yaml"
    ret = run_subprocess(["bioimageio", "update-format", str(in_path), str(path)])
    assert ret.returncode == 0
    assert path.exists()


def test_update_rdf(unet2d_nuclei_broad_base_path, tmp_path):
    in_path = unet2d_nuclei_broad_base_path / "rdf.yaml"
    assert in_path.exists()
    update_path = tmp_path / "update.yaml"
    assert yaml is not None
    yaml.dump(dict(name="updated", outputs=[{"name": "updated", "halo": ["KEEP", "DROP", 0, 9, 9]}]), update_path)
    out_path = tmp_path / "output.yaml"
    ret = run_subprocess(["bioimageio", "update-rdf", str(in_path), str(update_path), str(out_path)])
    assert ret.returncode == 0
    actual = yaml.load(out_path)
    assert actual["name"] == "updated"
    assert actual["outputs"][0]["name"] == "updated"
    assert actual["outputs"][0]["halo"] == [0, 0, 9, 9]



from inspect import getmembers

import pytest

from bioimageio.spec.shared import fields
from bioimageio import spec


@pytest.mark.parametrize(
    "schema_raw_nodes_pair",
    [
        (spec.model.schema, spec.model.raw_nodes),
        (spec.model.v0_3.schema, spec.model.v0_3.raw_nodes),
        (spec.model.v0_4.schema, spec.model.v0_4.raw_nodes),
        (spec.rdf.schema, spec.rdf.raw_nodes),
        (spec.rdf.v0_2.schema, spec.rdf.v0_2.raw_nodes),
    ],
)
def test_model_spec(schema_raw_nodes_pair):
    schema, raw_nodes = schema_raw_nodes_pair
    from bioimageio.spec.shared.schema import SharedBioImageIOSchema
    from bioimageio.spec.shared.raw_nodes import RawNode

    schema_names = {
        name for name, cls in getmembers(schema) if isinstance(cls, type) and issubclass(cls, SharedBioImageIOSchema)
    }
    # remove SharedBioImageIOSchema from schema names
    schema_names -= {SharedBioImageIOSchema.__name__}
    assert schema_names  # did we get any?

    node_names = {name for name, cls in getmembers(raw_nodes) if isinstance(cls, type) and issubclass(cls, RawNode)}
    # remove any node_names that are fields
    field_names = {
        name for name, cls in getmembers(fields) if isinstance(cls, type) and issubclass(cls, fields.DocumentedField)
    }
    assert field_names  # did we get any?
    node_names -= field_names
    # if present, ignore raw_nodes.ImportableModule and raw_nodes.ImportableSourceFile which are coming from
    # fields.ImportableSource
    node_names -= {n for n in {"ImportableModule", "ImportableSourceFile"} if hasattr(raw_nodes, n)}

    assert node_names  # did we get any?

    assert node_names, schema_names



from typing import Union

import pytest
from marshmallow import ValidationError
from pytest import raises


pre_and_post_processing = [
    (True, "binarize", {"threshold": 0.5}),
    (False, "binarize", {"mode": "fixed", "threshold": 0.5}),
    (True, "clip", {"min": 0.2, "max": 0.5}),
    (False, "clip", {"min": "min", "max": 0.5}),
    (True, "scale_linear", {"gain": 2, "offset": 0.5, "axes": "xy"}),
    (False, "scale_linear", {"gain": 2, "offset": 0.5, "axes": "b"}),
    (True, "sigmoid", {}),
    (False, "sigmoid", {"axes": "x"}),
    (True, "zero_mean_unit_variance", {"mode": "fixed", "mean": 1, "std": 2, "axes": "xy"}),
    (False, "zero_mean_unit_variance", {"mode": "unknown", "mean": 1, "std": 2, "axes": "xy"}),
    (True, "scale_range", {"mode": "per_sample", "axes": "xy"}),
    (False, "scale_range", {"mode": "fixed", "axes": "xy"}),
    (True, "scale_range", {"mode": "per_sample", "axes": "xy", "min_percentile": 5, "max_percentile": 50}),
    (False, "scale_range", {"mode": "per_sample", "axes": "xy", "min_percentile": 50, "max_percentile": 50}),
    (False, "scale_range", {"mode": "per_sample", "axes": "xy", "min": 0}),
]


@pytest.fixture(
    params=[
        ("Preprocessing", valid, name, kwargs)
        for valid, name, kwargs in pre_and_post_processing
        + [(False, "scale_range", {"mode": "per_dataset", "axes": "xy", "reference_tensor": "some_input_tensor_name"})]
    ]
    + [
        ("Postprocessing", valid, name, kwargs)
        for valid, name, kwargs in pre_and_post_processing
        + [
            (True, "scale_range", {"mode": "per_sample", "axes": "xy"}),
            (True, "scale_range", {"mode": "per_dataset", "axes": "xy", "reference_tensor": "some_input_tensor_name"}),
            (True, "scale_mean_variance", {"mode": "per_sample", "reference_tensor": "some_tensor_name"}),
            (False, "scale_mean_variance", {"mode": "per_sample"}),
            (True, "scale_mean_variance", {"mode": "per_dataset", "reference_tensor": "some_tensor_name"}),
            (False, "scale_mean_variance", {"mode": "per_dataset"}),
        ]
    ]
)
def processing_test_data(request):
    return request.param


def test_processing(processing_test_data):
    proc_class_name, valid, name, kwargs = processing_test_data
    from bioimageio.spec.model import schema

    proc: Union[schema.Preprocessing, schema.Postprocessing] = getattr(schema, proc_class_name)()

    if valid:
        proc.load({"name": name, "kwargs": kwargs})
    else:
        with raises(ValidationError):
            proc.load({"name": name, "kwargs": kwargs})



import copy
from typing import Any, Dict


def remove_slash_from_names(data: Dict[str, Any]) -> None:
    if "name" in data and isinstance(data["name"], str):
        data["name"] = data["name"].replace("/", "").replace("\\", "")

    # remove slashes in author/maintainer name
    authors = data.get("authors")
    maintainers = data.get("maintainers")
    persons = (authors if isinstance(authors, list) else []) + (maintainers if isinstance(maintainers, list) else [])
    for p in persons:
        if isinstance(p, dict) and "name" in p:
            p["name"] = p["name"].replace("/", "").replace("\\", "")


def maybe_convert(data: Dict[str, Any]) -> Dict[str, Any]:
    data = copy.deepcopy(data)

    # we unofficially accept strings as author entries...
    authors = data.get("authors")
    if isinstance(authors, list):
        data["authors"] = [{"name": a} if isinstance(a, str) else a for a in authors]

    if data.get("format_version") in ("0.2.0", "0.2.1"):
        data["format_version"] = "0.2.2"

    if data.get("format_version") == "0.2.2":
        remove_slash_from_names(data)
        data["format_version"] = "0.2.3"

    return data
